﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Security.Cryptography;
using System.Text;
using System.Xml;

namespace ABCommon
{
    public class CredentialStore
    {
        private static string CRED_FILENAME = "credentials.xml";

        private string account = null;
        private string password = null;
        private string filename = null;
        private string latestVersion = null;
        private string key = null;
        private bool doesMatchFile;

        #region Lifecycle methods

        public CredentialStore(string key, bool loadInitialValues)
        {
            this.key = key;

            if (loadInitialValues)
                LoadCredentials();
            else
                this.doesMatchFile = false;
        }

        ~CredentialStore()
        {
            if (!this.doesMatchFile)
                StoreCredentials();
        }
        #endregion

        #region Accessor methods

        public string GetAccount()
        {
            return this.account;
        }

        public void SetAccount(string value)
        {
            if ((value == null && this.account != null) ||
                !value.Equals(this.account))
            {
                this.account = value;
                this.doesMatchFile = false;
            }
        }

        public string GetPassword()
        {
            if (this.password == null)
                return null;
            else if (this.key != null)
                return Crypt(this.password, this.key, false);
            else
                return this.password;
        }

        public void SetPassword(string value)
        {
            string storedValue;
            if (key != null && value != null)
                storedValue = Crypt(value, this.key, true);
            else
                storedValue = value;

            if ((storedValue == null && this.password != null) ||
                !storedValue.Equals(this.password))
            {
                this.password = storedValue;
                this.doesMatchFile = false;
            }
        }

        public string GetFilename()
        {
            return this.filename;
        }

        public void SetFilename(string value)
        {
            if ((value == null && this.filename != null) ||
                !value.Equals(this.filename))
            {
                this.filename = value;
                this.doesMatchFile = false;
            }
        }

        public string GetLatestVersion()
        {
            return this.latestVersion;
        }

        public void SetLatestVersion(string value)
        {
            if ((value == null && this.latestVersion != null) ||
                !value.Equals(this.latestVersion))
            {
                this.latestVersion = value;
                this.doesMatchFile = false;
            }
        }

        public bool IsSynced()
        {
            return this.doesMatchFile;
        }
        #endregion

        #region Persistence methods

        public void LoadCredentials()
        {
            XmlDocument doc = new XmlDocument();
            try
            {
                doc.Load(CRED_FILENAME);

                XmlNodeList credList = doc.GetElementsByTagName("credential");

                // For now, assume only one entry
                if (credList == null || credList.Count < 1)
                    throw new IOException("Bad credential file format");

                XmlElement credential = (XmlElement)credList.Item(0);

                XmlNodeList nodeList = credential.GetElementsByTagName("account");
                if (nodeList != null && nodeList.Count > 0)
                    this.account = nodeList[0].InnerText;
                else
                    this.account = null;

                nodeList = credential.GetElementsByTagName("password");
                if (nodeList != null && nodeList.Count > 0)
                    this.password = nodeList[0].InnerText;
                else
                    this.password = null;

                nodeList = credential.GetElementsByTagName("filename");
                if (nodeList != null && nodeList.Count > 0)
                    this.filename = nodeList[0].InnerText;
                else
                    this.filename = null;

                nodeList = credential.GetElementsByTagName("latestVersion");
                if (nodeList != null && nodeList.Count > 0)
                    this.latestVersion = nodeList[0].InnerText;
                else
                    this.latestVersion = null;

                this.doesMatchFile = true;
            }
            catch (Exception e)
            {
                Console.WriteLine(e.Message);
            }
        }

        public void StoreCredentials()
        {
            if (!this.doesMatchFile)
            {
                try
                {
                    FileStream credFile = new FileStream(CRED_FILENAME, FileMode.Create, FileAccess.ReadWrite);
                    XmlTextWriter writer = new XmlTextWriter(credFile, null);
                    writer.Formatting = Formatting.Indented;

                    writer.WriteStartElement("credential");

                    if (this.account != null)
                    {
                        writer.WriteStartElement("account");
                        writer.WriteString(this.account);
                        writer.WriteEndElement();
                    }

                    if (this.password != null)
                    {
                        writer.WriteStartElement("password");
                        writer.WriteString(this.password);
                        writer.WriteEndElement();
                    }

                    if (this.filename != null)
                    {
                        writer.WriteStartElement("filename");
                        writer.WriteString(this.filename);
                        writer.WriteEndElement();
                    }

                    if (this.latestVersion != null)
                    {
                        writer.WriteStartElement("latestVersion");
                        writer.WriteString(this.latestVersion);
                        writer.WriteEndElement();
                    }

                    writer.WriteEndElement();

                    writer.Flush();
                    credFile.Close();
                }
                catch (Exception e)
                {
                    Console.WriteLine(e.Message);
                }
            }
        }
        #endregion

        #region Encryption/Decryption helper method

        public static string Crypt(string data, string password, bool isEncrypting) 
        { 
	        byte[] saltBytes = new byte[] {0x26, 0x19, 0x81, 0x4E, 0xA0, 0x6D, 0x95, 0x34, 0x26, 0x75, 0x64, 0x05, 0xF6}; 

	        PasswordDeriveBytes passwordBytes = new PasswordDeriveBytes(password, saltBytes);

	        Rijndael algorithm = Rijndael.Create();
	        algorithm.Key = passwordBytes.GetBytes(32);
	        algorithm.IV  = passwordBytes.GetBytes(16);

	        ICryptoTransform cryptoTransformer = (isEncrypting) ? algorithm.CreateEncryptor() : algorithm.CreateDecryptor();

	        MemoryStream memStream = new MemoryStream();
        	CryptoStream cryptoStream = new CryptoStream(memStream, cryptoTransformer, CryptoStreamMode.Write);

	        byte[] dataBytes;
        	if (isEncrypting)
                dataBytes = Encoding.Unicode.GetBytes(data);
	        else
                dataBytes = Convert.FromBase64String(data);

	        try
	        {
		        cryptoStream.Write(dataBytes, 0, dataBytes.Length);
		        cryptoStream.Close();
		        if (isEncrypting)
                    return Convert.ToBase64String(memStream.ToArray());
		        else
                    return Encoding.Unicode.GetString(memStream.ToArray());
	        }
	        catch
            {
                return null;
            }
        }
        #endregion
    }
}
