﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Threading;
using XihSolutions.DotMSN;
using XihSolutions.DotMSN.DataTransfer;

namespace ABCommon
{
    public delegate void sendMessageDelegate(string message, string excludedContact);

    public class MSNManager
    {
        public static string CLIENT_CODE = "Q1P7W2E4J9R8U3S5";

        private Messenger messenger;
        
        //these EventHandlers allow users of the MSNManager class to contribute their own behavior
        //on top of the generic behavior provided
        public event EventHandler transferInviteReceived;
        public event EventHandler contactStatusChanged;
        public event EventHandler signedInEvent;

        private CredentialStore store;
        private Delegate transferDelegate;
        private Delegate upViewerDelegate;
        private Delegate statusDelegate;
        private string revisionPath;
        private FileStream metaData;
        private int threadCount = 0;

        private Dictionary<string, ConversationHandler> conversations;
        private bool signedIn = false;
        private bool fullVersion = false;

        public MSNManager(CredentialStore store, Delegate transferDelegate, Delegate updateViewerDelegate, Delegate statusDelegate, bool fullVersion)
        {
            this.store = store;
            this.transferDelegate = transferDelegate;
            this.upViewerDelegate = updateViewerDelegate;
            this.statusDelegate = statusDelegate;
            this.fullVersion = fullVersion;

            messenger = new Messenger();
            refreshCredentials();
            
            //defining handlers for the various MSN related events that we are interested in
            messenger.NameserverProcessor.ConnectionEstablished += new EventHandler(NameserverProcessor_ConnectionEstablished);
            messenger.Nameserver.SignedIn += new EventHandler(Nameserver_SignedIn);
            messenger.ConversationCreated += new ConversationCreatedEventHandler(messenger_ConversationCreated);
            messenger.NameserverProcessor.ConnectingException += new XihSolutions.DotMSN.Core.ProcessorExceptionEventHandler(NameserverProcessor_ConnectingException);
            messenger.NameserverProcessor.ConnectionException += new XihSolutions.DotMSN.Core.ProcessorExceptionEventHandler(NameserverProcessor_ConnectionException);
            messenger.Nameserver.ServerErrorReceived += new XihSolutions.DotMSN.Core.ErrorReceivedEventHandler(Nameserver_ServerErrorReceived);
            messenger.Nameserver.AuthenticationError += new XihSolutions.DotMSN.Core.HandlerExceptionEventHandler(Nameserver_AuthenticationError);
            messenger.Nameserver.ContactRemoved += new ListMutatedAddedEventHandler(Nameserver_ContactRemoved);
            messenger.TransferInvitationReceived += new MSNSLPInvitationReceivedEventHandler(messenger_TransferInvitationReceived);
            messenger.Nameserver.ContactAdded += new ListMutatedAddedEventHandler(Nameserver_ContactAdded);
            messenger.Nameserver.SynchronizationCompleted += new EventHandler(Nameserver_SynchronizationCompleted);
            messenger.Nameserver.ContactStatusChanged += new ContactStatusChangedEventHandler(Nameserver_ContactStatusChanged);

            conversations = new Dictionary<string, ConversationHandler>();

            metaData = new FileStream("metadata.txt", FileMode.OpenOrCreate, FileAccess.ReadWrite);
            metaData.Seek(0, SeekOrigin.End);
        }

        public Messenger getMessenger()
        {
            return messenger;
        }

        public void refreshCredentials()
        {
            messenger.Credentials.Account = store.GetAccount();
            messenger.Credentials.Password = store.GetPassword();
            messenger.Credentials.ClientID = "msmsgs@msnmsgr.com";
            messenger.Credentials.ClientCode = CLIENT_CODE;
        }
        
        public Boolean getSignedIn()
        {
            return signedIn;
        }
        
        public void setRevisionPath(string path)
        {
            revisionPath = path;
        }
        
        public FileStream getMetaData()
        {
            return metaData;
        }
        
        void Nameserver_SynchronizationCompleted(object sender, EventArgs e)
        {
        }

        void Nameserver_ContactStatusChanged(object sender, ContactStatusChangeEventArgs e)
        {
            if (contactStatusChanged != null)
                contactStatusChanged(sender, e);
        }

        void NameserverProcessor_ConnectionEstablished(object sender, EventArgs e)
        {
            Console.WriteLine("Established");
        }

        void Nameserver_ContactAdded(object sender, ListMutateEventArgs e)
        {
            Console.WriteLine(e.AffectedList.ToString());
            if (e.AffectedList == MSNLists.ReverseList)
            {
                messenger.Nameserver.AddContactToList(e.Contact, MSNLists.AllowedList);
                messenger.Nameserver.AddContactToList(e.Contact, MSNLists.ForwardList);
            }
        }

        void messenger_TransferInvitationReceived(object sender, MSNSLPInvitationEventArgs e)
        {
            Console.WriteLine("Invite Received");

            if (transferInviteReceived != null)
                transferInviteReceived(sender, e);
        }

        void Nameserver_ContactRemoved(object sender, ListMutateEventArgs e)
        {
            Console.WriteLine("Contact removed");
        }

        void Nameserver_AuthenticationError(object sender, ExceptionEventArgs e)
        {
            Console.WriteLine(e.Exception.Message.ToString());
        }

        void Nameserver_ServerErrorReceived(object sender, MSNErrorEventArgs e)
        {
            Console.WriteLine(e.MSNError.ToString());
        }

        void NameserverProcessor_ConnectionException(object sender, ExceptionEventArgs e)
        {
            Console.WriteLine(e.Exception.Message.ToString());
        }

        void NameserverProcessor_ConnectingException(object sender, ExceptionEventArgs e)
        {
            Console.WriteLine(e.Exception.Message.ToString());
        }

        //called when a conversation is created, adds an entry to the list of known conversations
        void messenger_ConversationCreated(object sender, ConversationCreatedEventArgs e)
        {
            if (e.Initiator == null)
            {
                ConversationHandler ch = new ConversationHandler(e.Conversation,
                    transferDelegate,
                    upViewerDelegate,
                    new sendMessageDelegate(broadcastMessage),
                    statusDelegate,
                    metaData, revisionPath, fullVersion);

                foreach (DictionaryEntry d in e.Conversation.Switchboard.Contacts)
                {
                    Contact c = (Contact)d.Value;
                    conversations.Add(c.Mail, ch);
                }

                Thread oThread = new Thread(new ThreadStart(ch.run));

                // Start the thread
                oThread.Start();
                oThread.IsBackground = true;
                threadCount++;
                Console.WriteLine("Thread Count: " + threadCount.ToString());
            }
        }

        void Nameserver_SignedIn(object sender, EventArgs e)
        {
            messenger.Owner.Status = PresenceStatus.Hidden;
           
            Console.WriteLine("Signed in!");
            messenger.Nameserver.SynchronizeContactList();

            signedIn = true;
            
            //copy ones from reverse to forward and allowed
            foreach (Contact c in messenger.ContactList.Reverse)
            {
                messenger.Nameserver.AddContactToList(c, MSNLists.AllowedList);
                messenger.Nameserver.AddContactToList(c, MSNLists.ForwardList);
            }

            if (signedInEvent != null)
            {
                signedInEvent(sender, e);
            }
        }

        public void broadcastMessage(string message)
        {
            broadcastMessage(message, null);
        }

        private void broadcastMessage(string message, string excludedContact)
        {
            try
            {
                foreach (Contact c in messenger.ContactList.Forward)
                {
                    if (c.Online)
                    {
                        if (!conversations.ContainsKey(c.Mail))
                        {
                            Conversation convo = messenger.CreateConversation();

                            convo.Invite(c);

                            //we have to sleep until the the session is established.  I think under the hood, dotMSN
                            //is handling all the handshaking with the remote contact
                            while (!convo.Switchboard.IsSessionEstablished)
                            {
                                Thread.Sleep(50);
                                Console.WriteLine("Sleeping");
                            }

                            ConversationHandler ch = new ConversationHandler(convo,
                                transferDelegate,
                                upViewerDelegate,
                                new sendMessageDelegate(broadcastMessage),
                                statusDelegate,
                                metaData, revisionPath, fullVersion);

                            conversations.Add(c.Mail, ch);

                            Thread oThread = new Thread(new ThreadStart(ch.run));

                            // Start the thread
                            oThread.Start();

                            oThread.IsBackground = true;

                            threadCount++;
                            Console.WriteLine("Thread Count: " + threadCount.ToString());
                        }
                    }
                }

                //putting this sleep here seems to be necessary.  If it's not here,
                //then sometimes the message is sent and sometimes it's not
                Thread.Sleep(1000);

                foreach (ConversationHandler ch in conversations.Values)
                {
                    //simple case (broadcast to everyone)
                    if (excludedContact == null)
                        ch.sendMessage(message);
                    else
                    {
                        //less simple case (exclude one recipient)
                        Console.WriteLine(ch.remoteAccount);
                        if (!ch.remoteAccount.Equals(excludedContact))
                            ch.sendMessage(message);
                    }
                }
            }
            catch (Exception e)
            {
                Console.WriteLine(e.Message);
            }
        }

        public void connect()
        {
            messenger.Connect();
        }
    }
}
