using System;
using System.Drawing;
using System.IO;
using System.Runtime.InteropServices;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading;
using System.Windows.Forms;
using ABCommon;
using Algorithm.Diff;
using gma.System.Windows;
using ICSharpCode.SharpZipLib.Zip;
using XihSolutions.DotMSN;
using XihSolutions.DotMSN.DataTransfer;


namespace IMAwareness
{
    public delegate void handleTransferDelegate(Contact receiver, string revision);
    public delegate void updateViewerDelegate(string message);
    public delegate void statusOverrideDelegate(Contact receiver, bool overrideStatus);


    //main class in the program.  Handles pretty much everything but the conversation stuff
    public partial class ABForm : Form
    {
        private string name;
        private string originalFilename;
        private string currentFilename;
        private string revisionPath;
        private string latestVersion;
        private int revisionCount = 1;

        private bool isOpen;
        private bool firstEdit = false;
        private bool statusOverridden = false;

        private int timeCount = 0;
        private int lastVersionTime = 0;
        private int lastRetrievalTime = 0;

        private CredentialStore store;
        private Messenger messenger;
        private MSNManager msnManager;

        private FileStream upLoad;

        private MainMenu mainMenu;
        private AccountForm accountForm;
        private BuddyForm buddyForm;

        private const int ICON_BUSY = 4;
        private const int ICON_EMPTY = 3;
        private const int ICON_IDLE = 1;
        private const int ICON_OFFLINE = 0;
        private const int ICON_ONLINE = 2;

        [DllImport("user32.dll")]
        static extern int GetForegroundWindow();

        [DllImport("user32.dll")]
        static extern int GetWindowText(int hWnd, StringBuilder text, int count);
        
        private UserActivityHook globalHook;

        #region Application form lifecycle methods

        public ABForm()
        {
            InitializeComponent();

            mainMenu = new MainMenu();
            MenuItem File = mainMenu.MenuItems.Add("&Properties");
            File.MenuItems.Add(new MenuItem("&Account", new EventHandler(this.AccountMenu_clicked)));
            File.MenuItems.Add(new MenuItem("Add &Buddy", new EventHandler(this.BuddyMenu_clicked)));
            File.MenuItems.Add(new MenuItem("-"));
            File.MenuItems.Add(new MenuItem("&Stop Sharing Artifact", new EventHandler(this.StopSharingMenu_clicked)));
            File.MenuItems.Add(new MenuItem("E&xit", new EventHandler(this.ExitMenu_clicked)));
            this.Menu = mainMenu;

            //hook global events. This is used to detect if a user is editing the file
            globalHook = new UserActivityHook(); // crate an instance
            globalHook.KeyPress += new KeyPressEventHandler(globalHook_KeyPressed);

            //get a credential store
            store = new CredentialStore(MSNManager.CLIENT_CODE, true);

            //set up the link to msn
            msnManager = new MSNManager(store,
                new handleTransferDelegate(handleTransfer), 
                new updateViewerDelegate(updateChatLog),
                new statusOverrideDelegate(statusOverride),
                true);

            msnManager.transferInviteReceived += new EventHandler(msnManager_TransferInvitationReceived);
            msnManager.contactStatusChanged += new EventHandler(msnManager_ContactStatusChanged);
            msnManager.signedInEvent += new EventHandler(msnManager_SignedIn);
            isOpen = false;

            messenger = msnManager.getMessenger();
             
            buddyForm = new BuddyForm();
            accountForm = new AccountForm(messenger.Credentials.Account, messenger.Credentials.Password);
        }

        private void ABForm_Load(object sender, EventArgs e)
        {
            //if we're restarting from a previous state, then go ahead and resume that
            string tempFilename = store.GetFilename();
            string tempLatest = store.GetLatestVersion();
            if (tempFilename != null && tempLatest != null)
            {
                startSharing(tempFilename, false);
                latestVersion = tempLatest;
            }
            else
            {
                Image img = filePreview.Image;
                filePreview.Image = img.GetThumbnailImage(this.filePreview.Size.Width,
                    this.filePreview.Size.Height,
                    new System.Drawing.Image.GetThumbnailImageAbort(img_GetThumbnailAborted),
                    IntPtr.Zero);
            }
        }

        private void ABForm_FormClosed(object sender, FormClosedEventArgs e)
        {
            msnManager.getMetaData().Close();
        }
        #endregion

        #region UI callback handlers

        private void AccountMenu_clicked(object sender, EventArgs e)
        {
            if (accountForm.IsDisposed)
                accountForm = new AccountForm(messenger.Credentials.Account, messenger.Credentials.Password);
            accountForm.Show();
            accountForm.FormClosed += new FormClosedEventHandler(credForm_FormClosed);
        }

        private void BuddyMenu_clicked(object sender, EventArgs e)
        {
            if (buddyForm.IsDisposed)
                buddyForm = new BuddyForm();
            buddyForm.Show();
            buddyForm.FormClosed += new FormClosedEventHandler(buddyForm_FormClosed);
        }

        private void StopSharingMenu_clicked(object sender, EventArgs e)
        {
            if (MessageBox.Show("Are you sure you want to stop sharing " + name + "?",
                                "Close Artifact",
                                MessageBoxButtons.YesNo) == DialogResult.Yes)
            {
                if (MessageBox.Show("Would you like to archive the working copy of " + name + "?",
                                    "Close Artifact",
                                    MessageBoxButtons.YesNo) == DialogResult.Yes)
                {
                    SaveFileDialog dlg = new SaveFileDialog();
                    dlg.FileName = name;
                    dlg.InitialDirectory = originalFilename.Substring(0, originalFilename.LastIndexOf(Path.DirectorySeparatorChar));
                    dlg.Title = "Archive Working Copy of " + name;
                    DialogResult saveResult = dlg.ShowDialog();

                    // The save dialog messes up our default path, so reset it before continuing
                    // (it screws up all of our existing paths if we don't)
                    Directory.SetCurrentDirectory(Application.StartupPath);

                    if (saveResult == DialogResult.OK)
                    {
                        try
                        {
                            File.Copy(currentFilename, dlg.FileName);
                        }
                        catch (Exception e2)
                        {
                            Console.WriteLine(e2.Message);
                            return;
                        }
                    }
                    else
                    {
                        return;
                    }
                }

                // Clear the credential store
                store.SetFilename(null);
                store.SetLatestVersion(null);
                store.StoreCredentials();

                // Delete the superfluous files
                msnManager.getMetaData().Close();
                File.Delete("metadata.txt");
                Directory.Delete(name.Substring(0, name.IndexOf(new FileInfo(originalFilename).Extension)), true);

                // Re-start the app
                Application.Exit();
                System.Diagnostics.Process.Start(Application.ExecutablePath);
            }
        }

        private void ExitMenu_clicked(object sender, EventArgs e)
        {
            Application.Exit();
        }

        void credForm_FormClosed(object sender, FormClosedEventArgs e)
        {
            string account = accountForm.email;
            string password = accountForm.password;

            store.SetAccount(account);
            store.SetPassword(password);
            store.StoreCredentials();
            msnManager.refreshCredentials();

            if (account != null &&
                !account.Equals("") &&
                password != null &&
                !password.Equals("") &&
                name != null)
            {
                messenger.Connect();
            }
        }

        void buddyForm_FormClosed(object sender, FormClosedEventArgs e)
        {
            string email = buddyForm.email;
            if (!email.Equals(""))
            {
                Contact c = (messenger.Nameserver.ContactList.GetContact(email));
                messenger.Nameserver.AddContactToList(c, MSNLists.AllowedList);
                messenger.Nameserver.AddContactToList(c, MSNLists.ForwardList);
            }
        }

        private void versionViewer_NodeMouseDoubleClick(object sender, TreeNodeMouseClickEventArgs e)
        {
            TreeNode selectedNode = e.Node;
            object tag = selectedNode.Tag;
            if (tag != null)
            {
                string path = selectedNode.Tag.ToString();
                System.Diagnostics.Process.Start(path);
            }
        }

        private void updateButton_Click(object sender, EventArgs e)
        {
            messenger.Owner.Name = name + "-" + "New Version Available";
            artifactStatusLabel.Text = "New Version Available";

            string newVersionPath = revisionPath + revisionCount++ + "-" + Environment.UserName + "-" + name;
            string old = "";
            FileInfo[] files = new DirectoryInfo(revisionPath).GetFiles();
            int revisionNum = (revisionCount - 2);
            string revision = revisionNum.ToString();
            for (int i = 0; i < files.Length; i++)
            {
                if (files[i].Name.StartsWith(revision))
                {
                    old = files[i].FullName;
                }
            }

            latestVersion = newVersionPath;
            store.SetLatestVersion(latestVersion);
            lastVersionTime = Environment.TickCount;
            //move the latest version into the 'files' directory
            File.Copy(originalFilename, currentFilename, true);
            //handle 'version control'- the file is modified by appending a number to its filename
            File.Copy(currentFilename, newVersionPath, true);

            compareAndSend(old, latestVersion);
            firstEdit = false;
        }

        private void sendButton_Click(object sender, EventArgs e)
        {
            updateChatLog("I say: \r\n" + msgSendBox.Text + "\r\n");
            msnManager.broadcastMessage(msgSendBox.Text);
        }

        private void fileOpenTimer_Tick(object sender, EventArgs e)
        {
            try
            {
                if (currentFilename != null)
                {
                    FileStream stream = System.IO.File.Open(originalFilename, FileMode.OpenOrCreate, FileAccess.Read, FileShare.None);
                    stream.Close();
                    isOpen = false;
                    firstEdit = false;

                    // Only update the status if it's not currently being overridden by someone else
                    if (!statusOverridden)
                    {
                        messenger.Owner.Status = PresenceStatus.Away;
                        messenger.Owner.Name = name + " is closed";
                        updateArtifactStatusIcon(ICON_IDLE);
                        artifactStatusLabel.Text = name + " is closed";
                    }
                }
            }
            catch
            {
                if (!isOpen)
                {
                    // Only update the status if it's not currently being overridden by someone else
                    if (msnManager.getSignedIn() && !statusOverridden)
                    {
                        messenger.Owner.Status = PresenceStatus.Online;
                        messenger.Owner.Name = Environment.UserName + " opened " + name;
                        updateArtifactStatusIcon(ICON_ONLINE);
                        artifactStatusLabel.Text = Environment.UserName + " opened " + name;

                        isOpen = true;
                    }
                }
            }
        }

        private void activeEditTimer_Tick(object sender, EventArgs e)
        {
            //don't bother doing anything if the file isn't open
            if (isOpen)
            {
                //check the difference between the current ticks and the last edit

                int diff = Environment.TickCount - timeCount;
                //if time count = 0, that means the file has been opened but never edited on
                if (diff > activeEditTimer.Interval || timeCount == 0)
                {
                    firstEdit = false;

                    // Only update the status if it's not currently being overridden by someone else
                    if (!statusOverridden)
                    {
                        messenger.Owner.Status = PresenceStatus.Idle;
                        messenger.Owner.Name = Environment.UserName + " paused editing on " + name;
                        updateArtifactStatusIcon(ICON_IDLE);
                        artifactStatusLabel.Text = Environment.UserName + " paused editing on " + name;
                    }
                }
            }
        }

        private void file_DragEnter(object sender, DragEventArgs e)
        {
            // make sure they're actually dropping files (not text or anything else)
            if (e.Data.GetDataPresent(DataFormats.FileDrop, false) == true)
                // allow them to continue
                // (without this, the cursor stays a "NO" symbol
                e.Effect = DragDropEffects.All;
        }

        private void file_DragDrop(object sender, DragEventArgs e)
        {
            Array a = (Array)e.Data.GetData(DataFormats.FileDrop);

            if (a != null)
            {
                // Extract string from first array element
                // (ignore all files except first if number of files are dropped).
                string s = a.GetValue(0).ToString();

                startSharing(s, true);
            }
        }

        public bool img_GetThumbnailAborted()
        {
            return true;
        }
        #endregion

        #region UI update helper methods

        //updates the form title
        private void updateABFormTitle()
        {
            this.Invoke(new updateABFormTitleDelegate(updateABFormTitleHelper), new object[] { });
        }

        private delegate void updateABFormTitleDelegate();
        private void updateABFormTitleHelper()
        {
            this.Text = messenger.Owner.Mail;
        }

        //updates the buddy icon for the artifact
        //index is the index into the image list
        private void updateArtifactStatusIcon(int index)
        {
            this.Invoke(new updateArtifactStatusIconDelegate(updateArtifactStatusIconHelper), new object[] { index });
        }

        private delegate void updateArtifactStatusIconDelegate(int i);
        private void updateArtifactStatusIconHelper(int index)
        {
            artifactStatusIcon.Image = buddyIconList.Images[index];
        }

        //updates the status text
        private void updateArtifactStatusLabel(string status)
        {
            this.Invoke(new updateArtifactStatusLabelDelegate(updateArtifactStatusLabelHelper), new object[] { status });
        }

        private delegate void updateArtifactStatusLabelDelegate(string s);
        private void updateArtifactStatusLabelHelper(string status)
        {
            this.artifactStatusLabel.Text = status;
        }

        //updates the version tree
        private void updateVersionTree()
        {
            this.Invoke(new updateVersionTreeDelegate(updateVersionTreeHelper), new object[] { });
        }

        private delegate void updateVersionTreeDelegate();
        private void updateVersionTreeHelper()
        {
            //full paths are stored in the 'Tag' value of a TreeNode
            TreeNode[] nodes = versionViewer.Nodes.Find("wcRoot", true);
            TreeNode root = nodes[0];
            root.Nodes.Clear();
            TreeNode node = new TreeNode(originalFilename);
            node.Tag = originalFilename;
            root.Nodes.Add(node);

            nodes = versionViewer.Nodes.Find("versionsRoot", true);
            root = nodes[0];
            root.Nodes.Clear();

            FileInfo[] files = new DirectoryInfo(revisionPath).GetFiles();

            int verCount = 0;

            for (int i = 0; i < files.Length; i++)
            {
                string filename = files[i].FullName;
                string[] tokens = files[i].Name.Split(new char[] { '-' });
                string revNumber = tokens[0];
                string userName = tokens[1];

                TreeNode temp = new TreeNode(revNumber + " (" + userName + ")");
                temp.Tag = filename;
                temp.Name = revNumber;

                Regex numandLetterPattern = new Regex("[0-9]*[a-z]");
                //is a child version
                if (numandLetterPattern.IsMatch(revNumber))
                {
                    Regex numPattern = new Regex("[0-9]*");
                    MatchCollection mCollection = numPattern.Matches(revNumber);
                    String val = mCollection[0].Value;
                    TreeNode[] parentArr = versionViewer.Nodes.Find(val, true);
                    TreeNode parent = parentArr[0];
                    parent.Nodes.Add(temp);
                }
                else
                {
                    root.Nodes.Add(temp);
                    verCount++;
                }
            }

            root.Text = "Versions (" + verCount + ")";
            versionViewer.Sort();
        }

        //updates the contact tree
        private void updateContactTree()
        {
            this.Invoke(new updateContactTreeDelegate(updateContactTreeHelper), new object[] { });
        }

        private delegate void updateContactTreeDelegate();
        private void updateContactTreeHelper()
        {
            TreeNode[] nodes = contactViewer.Nodes.Find("rootNode", true);
            TreeNode root = nodes[0];
            root.Nodes.Clear();
            int count = 0;

            foreach (Contact c in messenger.ContactList.All)
            {
                TreeNode buddy = new TreeNode(c.Name);

                if (c.Status == PresenceStatus.Offline)
                    buddy.ImageIndex = 0;
                else if (c.Status == PresenceStatus.Away || c.Status == PresenceStatus.Idle
                    || c.Status == PresenceStatus.Lunch || c.Status == PresenceStatus.Phone)
                    buddy.ImageIndex = 1;
                else if (c.Status == PresenceStatus.Online)
                    buddy.ImageIndex = 2;
                else if (c.Status == PresenceStatus.Busy)
                    buddy.ImageIndex = 4;
                else
                    buddy.ImageIndex = 3;

                root.Nodes.Add(buddy);
                count++;
            }

            root.Text = "Buddies (" + count + ")";
        }

        //updates the chat log
        private void updateChatLog(string message)
        {
            this.Invoke(new updateChatLogDelegate(updateChatLogHelper), new object[] { message });
        }

        private delegate void updateChatLogDelegate(string text);
        private void updateChatLogHelper(string text)
        {
            msgReceivedBox.AppendText(text + "\r\n");
        }
        #endregion

        #region Windows hooks callback handler

        //used to monitor if the user is editing the shared file
        public void globalHook_KeyPressed(object sender, KeyPressEventArgs e)
        {
            if (isDocumentActive()) {
                // Local edits always override externally-overridden status messages
                if (statusOverridden)
                    statusOverridden = false;

                if (!firstEdit && msnManager.getSignedIn())
                {
                    firstEdit = true;

                    messenger.Owner.Status = PresenceStatus.Online;
                    messenger.Owner.Name = Environment.UserName + " is editing " + name;
                    updateArtifactStatusIcon(ICON_ONLINE);
                    artifactStatusLabel.Text = Environment.UserName + " is editing " + name;
                }

                // Console.WriteLine("Editing on: " + file);
                timeCount = Environment.TickCount;
            }
        }
        #endregion

        #region Messenger callback handlers

        void msnManager_SignedIn(object sender, EventArgs e)
        {
            messenger.Owner.Status = PresenceStatus.Hidden;
            messenger.Owner.Name = name;
            updateArtifactStatusIcon(ICON_OFFLINE);
            updateArtifactStatusLabel("");

            updateABFormTitle();
        }

        void msnManager_ContactStatusChanged(object sender, EventArgs e)
        {
            //update contact tree viewer
            updateContactTree();
        }

        void msnManager_TransferInvitationReceived(object sender, EventArgs args)
        {
            MSNSLPInvitationEventArgs e = (MSNSLPInvitationEventArgs)args;

            Console.WriteLine("Invite Received");

            MSNSLPHandler msnHandler = (MSNSLPHandler)sender;
            string account = msnHandler.MessageSession.RemoteContact;

            Contact c = messenger.Nameserver.ContactList.GetContact(account);
            string senderName = c.Name;
            string filename = "";
            string version;

            //if there hasn't been a version submitted since the last time a copy was retrieved
            //or if the file being sent is a copy of the working copy, create a new version
            if ((lastRetrievalTime > lastVersionTime) || e.Filename.Split('-')[0].Equals("wc") || (lastRetrievalTime == 0))
            {
                filename = revisionPath + revisionCount++ + "-" + senderName + "-" + name;
                int tempInt = (revisionCount - 1);
                version = tempInt.ToString();
            }
            //else the file being sent is a subversion, so find out what the approriate letter will be
            else
            {
                String[] temp = e.Filename.Split('-');
                string prevRevision = temp[0];
                Regex numandLetterPattern = new Regex("[0-9]*[a-z]");
                //checks to see if the received version was already a child version
                //if no, just add an 'a'
                if (!numandLetterPattern.IsMatch(prevRevision))
                {
                    filename = revisionPath + temp[0] + "a" + "-" + senderName + "-" + name;
                    version = temp[0] + "a";
                }
                //find out what letter is attached to the received version
                else
                {
                    Regex charPattern = new Regex("[a-z]");
                    MatchCollection mCollection = charPattern.Matches(prevRevision);
                    String val = mCollection[0].Value;
                    char oldVersion = val[0];
                    char newVersion = oldVersion;
                    newVersion++;
                    prevRevision = prevRevision.Replace(oldVersion, newVersion);
                    filename = revisionPath + prevRevision + "-" + senderName + "-" + name;
                    version = prevRevision;
                }
            }

            latestVersion = filename;
            store.SetLatestVersion(latestVersion);
            lastVersionTime = Environment.TickCount;
            e.TransferSession.DataStream = new FileStream(filename, FileMode.Create, FileAccess.Write);
            e.Accept = true;

            e.TransferSession.TransferFinished += new EventHandler(TransferSession_TransferFinished);

            String time = DateTime.Now.ToString("ddd, hh:mmtt");
            updateChatLog(c.Name + " says:\r\n" + " " + time + " " + c.Name + " is sending version " + version + "..." + new FileInfo(filename).Name + " sent");

            updateVersionTree();
        }
        #endregion

        #region TransferSession callback handlers

        void TransferSession_TransferFinished(object sender, EventArgs e)
        {
            P2PTransferSession session = (P2PTransferSession)sender;
            
            FileStream fileStream = (FileStream)session.DataStream;
            
            fileStream.Close();

            string oldVersion="";
            FileInfo[] files = new DirectoryInfo(revisionPath).GetFiles();
            int lastRevision = revisionCount -1;
            
            string revision = lastRevision.ToString();

            Console.WriteLine(revision);
            for (int i = 0; i < files.Length; i++)
            {
                if (files[i].Name.StartsWith(revision))
                {
                    oldVersion = files[i].FullName;
                }
            }

            firstEdit = false;
            compareAndSend(oldVersion, latestVersion);
            
            Console.WriteLine("Receiving Xfer done");
        }
        #endregion

        #region P2PTransferSession callback handlers

        void session_TransferStarted(object sender, EventArgs e)
        {
            Console.WriteLine("Sending Xfer started");
        }

        void session_TransferFinished(object sender, EventArgs e)
        {
            Console.WriteLine("Sending Xfer finished");
            upLoad.Close();

            //clean up
            File.Delete("temp");
            File.Delete("wc_temp");
            if (File.Exists("ArtifactBuddy-" + name + ".zip"))
            {
                File.Delete("ArtifactBuddy-" + name + ".zip");
            }
        }
        #endregion

        #region ABForm delegate handlers

        private delegate void bringToFrontDelegate();
        private void handleTransfer(Contact c, string revision)
        {
            try
            {
                string tempFilename="";
                string suggestedFilename = "";
                Console.WriteLine(revision);
              
                //find the file that the user wants
                FileInfo[] files=new DirectoryInfo(revisionPath).GetFiles();

                if (revision.Equals("LATEST"))
                {
                    tempFilename = latestVersion;
                    suggestedFilename = (revisionCount - 1)+"-"+name ;
                    lastRetrievalTime = Environment.TickCount;
                    String time = DateTime.Now.ToString("ddd, hh:mmtt");
                    updateChatLog(c.Name + " says:\r\n" + " " +time+" GET LATEST... version " + (revisionCount-1) + " retrieved by " + c.Name);
                }
                else if(revision.Equals("TRANSCRIPT"))
                {
                    File.Copy(msnManager.getMetaData().Name, "temp");
                    tempFilename = "temp";
                    suggestedFilename = "transcript.xml";
                    String time = DateTime.Now.ToString("ddd, hh:mmtt");
                    updateChatLog(c.Name + " says:\r\n" + " " + time + " GET TRANSCRIPT...  retrieved by " + c.Name);
                }
                else if (revision.Equals("WORKING"))
                {
                    //copy original to temp file
                    File.Copy(originalFilename, "wc_temp");
                    tempFilename = "wc_temp";
                    suggestedFilename = "wc" + "-" + name;
                    String time = DateTime.Now.ToString("ddd, hh:mmtt");
                    updateChatLog(c.Name + " says:\r\n" + " " + time + " GET WORKING...  retrieved by " + c.Name);
                }
                else if (revision.Equals("ORIGINAL"))
                {
                    for (int i = 0; i < files.Length; i++)
                    {
                        if (files[i].Name.ToUpper().StartsWith("1-"))
                        {
                            tempFilename = files[i].FullName;
                            suggestedFilename = "1" + "-" + name;
                            lastRetrievalTime = Environment.TickCount;
                            String time = DateTime.Now.ToString("ddd, hh:mmtt");
                            updateChatLog(c.Name + " says:\r\n" + " " + time + " GET ORIGINAL... version 1 retrieved by " + c.Name);
                            break;
                        }
                    }
                }
                else if (revision.Equals("CONTROL"))
                {
                    // Make sure the dialog doesn't get lost in the shuffle
                    this.Invoke(new bringToFrontDelegate(BringToFront), new object[] { });
                    if (MessageBox.Show(c.Name + " would like to take control of the artifact named " + name + ".\n\nWould you like to approve this request?",
                                        "Request Requires Approval", MessageBoxButtons.YesNo, MessageBoxIcon.Exclamation) == DialogResult.No)
                    {
                        msnManager.broadcastMessage(c.Name + " requested control.\r\n  Transfer of control was declined.");
                        return;
                    }

                    tempFilename = "ArtifactBuddy-" + name + ".zip";
                    suggestedFilename = tempFilename;

                    // Create a zip with everything we need
                    // First, check that we have an existing ZIP of the app
                    if (!File.Exists("ABFull.zip"))
                        Console.WriteLine("Error -- Artfact Buddy installation appears to be damaged. Please re-install.");
                    else
                    {
                        // Next, construct a self-contained app bundle:
                        // - Decompress the compressed app into a temp directory
                        // - Copy the original file into the directory (for easy editing)
                        // - Modify the credentials to point to this (relative)
                        //   path as the source file
                        // - Copy over the rest of our housekeeping metadata
                        // - Copy our version archive
                        // - Copy over the original ZIP file, itself, so that this
                        //   derivative can also pass itself along
                        // - ZIP it all back up
                        // - Clean up the mess
                        string tempDir = getTempDirectory();
                        try
                        {
                            Console.WriteLine("temp dir=" + tempDir);
                            FastZip zip = new FastZip();
                            zip.ExtractZip("ABFull.zip", tempDir, "");

                            File.Copy(originalFilename, tempDir + Path.DirectorySeparatorChar + name, true);

                            store.SetFilename(name);
                            store.StoreCredentials();

                            File.Copy("credentials.xml", tempDir + Path.DirectorySeparatorChar + "credentials.xml", true);
                            File.Copy("metadata.txt", tempDir + Path.DirectorySeparatorChar + "metadata.txt", true);

                            string directory = name.Substring(0, name.IndexOf(new FileInfo(originalFilename).Extension));
                            copyDirectory(directory, tempDir + Path.DirectorySeparatorChar + directory);

                            File.Copy("ABFull.zip", tempDir + Path.DirectorySeparatorChar + "ABFull.zip", true);

                            zip.CreateZip(tempFilename, tempDir, true, "");
                        }
                        catch (Exception e)
                        {
                            Console.WriteLine(e.Message);
                        }
                        finally
                        {
                            // Point back to the original, original filename
                            store.SetFilename(originalFilename);
                            store.StoreCredentials();

                            // Tidy up the temp folder
                            Directory.Delete(tempDir, true);
                        }
                    }

                    msnManager.broadcastMessage(c.Name + " requested control.\r\n  Control is now being transferred");
                }
                else
                {
                    for (int i = 0; i < files.Length; i++)
                    {

                        if (files[i].Name.ToUpper().StartsWith(revision.ToUpper()))
                        {
                            tempFilename = files[i].FullName;
                            suggestedFilename = revision + "-" + name;
                            lastRetrievalTime = Environment.TickCount;
                            String time = DateTime.Now.ToString("ddd, hh:mmtt");
                            updateChatLog(c.Name + " says:\r\n" + " " + time + " GET ... version " + revision + " retrieved by " + c.Name);
                            break;
                        }
                    }
                }

                if (!tempFilename.Equals(""))
                {
                    MSNSLPHandler msnslpHandler = messenger.GetMSNSLPHandler(c.Mail);
                    upLoad = new FileStream(tempFilename, FileMode.Open, FileAccess.Read, FileShare.Read); //currentFilename
                    P2PTransferSession session = msnslpHandler.SendInvitation(messenger.Owner.Mail, c.Mail, Path.GetFileName(suggestedFilename), upLoad);

                    session.TransferStarted += new EventHandler(session_TransferStarted);
                    session.TransferFinished += new EventHandler(session_TransferFinished);
                }
                else
                {
                    msnManager.broadcastMessage(c.Name + " says:\r\n" + " GET ... version " + revision + " (Error: Revision not found!)");
                }
            }
            catch (Exception e)
            {
                Console.WriteLine(e.Message);
            }
        }

        private void statusOverride(Contact c, bool overrideStatus)
        {
            statusOverridden = overrideStatus;
            String time = DateTime.Now.ToString("ddd, hh:mmtt");

            if (statusOverridden)
            {
                messenger.Owner.Status = PresenceStatus.Online;
                messenger.Owner.Name = c.Name + " is editing " + name;
                updateArtifactStatusIcon(ICON_ONLINE);
                updateArtifactStatusLabel(c.Name + " is editing " + name);
                
                updateChatLog(c.Name + " says:\r\n" + " " + time + " ANNOUNCE EDITING");
            }
            else
            {
                messenger.Owner.Status = PresenceStatus.Away;
                messenger.Owner.Name = c.Name + " has finished editing " + name;
                updateArtifactStatusIcon(ICON_IDLE);
                updateArtifactStatusLabel(c.Name + " has finished editing " + name);

                updateChatLog(c.Name + " says:\r\n" + " " + time + " ANNOUNCE EDIT DONE");

                isOpen = false;
                firstEdit = false;
            }
        }
        #endregion

        #region Miscellaneous helper methods

        private void startSharing(string filename, bool createWorkingCopy)
        {
            //the original filename is saved and NOT modified at all
            originalFilename = filename;
            store.SetFilename(filename);

            name = new FileInfo(originalFilename).Name;

            //currentFilename is a copy of the file that is currently being edited, the working copy
            string directory = name.Substring(0, name.IndexOf(new FileInfo(originalFilename).Extension));
            revisionPath = directory + "//versions//";
            currentFilename = directory + "//" + name;
            msnManager.setRevisionPath(revisionPath);

            if (!Directory.Exists(directory))
                Directory.CreateDirectory(directory);

            if (!Directory.Exists(revisionPath))
                Directory.CreateDirectory(revisionPath);

            if (createWorkingCopy)
            {
                //the file is copied to the files directory.  
                try
                {
                    File.Copy(originalFilename, currentFilename, true);
                    string newVersionPath = revisionPath + revisionCount++ + "-" + Environment.UserName + "-" + name;
                    latestVersion = newVersionPath;
                    store.SetLatestVersion(latestVersion);
                    //automatically create the first version when a file is dropped in
                    File.Copy(originalFilename, newVersionPath, true);
                }
                catch (Exception exp)
                {
                    Console.WriteLine(exp.Message);
                }

            }

            artifactStatusText.Text = name;

            updateVersionTree();

            Image img;
            //try to automatically grab the thumbnail for the file, via the Windows Shell
            //if that doesn't work, open the file and grab a screen shot of it
            try
            {
                vbAccelerator.Components.Shell.ThumbnailCreator tc = new vbAccelerator.Components.Shell.ThumbnailCreator();
                tc.DesiredSize = filePreview.Size;
                img = tc.GetThumbNail(originalFilename);
                filePreview.Image = img.GetThumbnailImage(this.filePreview.Size.Width, this.filePreview.Size.Height,
                new System.Drawing.Image.GetThumbnailImageAbort(img_GetThumbnailAborted), IntPtr.Zero);
            }
            catch
            {
                System.Diagnostics.Process p = System.Diagnostics.Process.Start(originalFilename);

                string title = getActiveWindowTitle();
                Console.WriteLine(title);
                Console.WriteLine(this.Text);

                while ((title != null && !title.Contains(name.Substring(0, name.IndexOf('.')))) || title.Equals(this.Text))
                {
                    Console.WriteLine(title);
                    Console.WriteLine(name.Substring(0, name.IndexOf('.')));
                    Thread.Sleep(20);

                    title = getActiveWindowTitle();
                }

                DesktopLib.WindowCapture capture = new DesktopLib.WindowCapture(DesktopLib.Win32.GetActiveWindowHandle());
                img = capture.GrabWindow();

                filePreview.Image = img.GetThumbnailImage(this.filePreview.Size.Width, this.filePreview.Size.Height,
                new System.Drawing.Image.GetThumbnailImageAbort(img_GetThumbnailAborted), IntPtr.Zero);

                img = filePreview.Image;

                //crops the screenshot to remove the task bar and applications menus
                //create the destination (cropped) bitmap
                Bitmap bmpCropped = new Bitmap(img.Height, img.Width);
                //create the graphics object to draw with
                Graphics g = Graphics.FromImage(bmpCropped);

                Rectangle rectDestination = new Rectangle(0, 0, bmpCropped.Width, bmpCropped.Height);
                Rectangle rectCropArea = new Rectangle(20, 20, img.Width - 35, img.Height - 50);

                //draw the rectCropArea of the original image to the rectDestination of bmpCropped
                g.DrawImage(img, rectDestination, rectCropArea, GraphicsUnit.Pixel);
                //release system resources
                g.Dispose();

                filePreview.Image = bmpCropped;
            }

            //attempt to use thumbnail as artifact buddy's display pic
            DisplayImage displayImage = new DisplayImage();
            displayImage.Image = filePreview.Image;
            messenger.Owner.DisplayImage = displayImage;
            messenger.Owner.BroadcastDisplayImage();

            if (!messenger.Credentials.Account.Equals("") && !messenger.Credentials.Password.Equals(""))
            {
                messenger.Connect();
            }

            revisionCount = determineRevision();
        }

        private int determineRevision()
        {
            FileInfo[] files = new DirectoryInfo(revisionPath).GetFiles();
            int highestRevision = 0;

            for (int i = 0; i < files.Length; i++)
            {
                string filename = files[i].FullName;
                string[] tokens = files[i].Name.Split(new char[] { '-' });
                string revNumber = tokens[0];
                int temp;
                Regex numandLetterPattern = new Regex("[0-9]*[a-z]");

                //is a child version
                if (numandLetterPattern.IsMatch(revNumber))
                {
                    Regex numPattern = new Regex("[0-9]*");
                    MatchCollection mCollection = numPattern.Matches(revNumber);
                    String val = mCollection[0].Value;
                    temp = System.Convert.ToInt32(val, 10);
                }
                else
                {
                    temp = System.Convert.ToInt32(revNumber, 10);
                }

                if (temp > highestRevision)
                    highestRevision = temp;
            }

            return (highestRevision + 1);
        }

        private string getTempDirectory()
        {
            string path = Path.GetRandomFileName();
            path = Path.Combine(Path.GetTempPath(), path);
            Directory.CreateDirectory(path);
            return path;
        }

        public void copyDirectory(string source, string destination){
            String[] files;

            if (destination[destination.Length - 1] != Path.DirectorySeparatorChar) 
                destination += Path.DirectorySeparatorChar;
            if (!Directory.Exists(destination))
                Directory.CreateDirectory(destination);
            files = Directory.GetFileSystemEntries(source);
            foreach (string file in files){
                // Subdirectories (recursion)
                if (Directory.Exists(file)) 
                    copyDirectory(file, destination + Path.GetFileName(file));
                // Files in directory
                else 
                    File.Copy(file, destination + Path.GetFileName(file), true);
            }
        }

        private string getActiveWindowTitle()
        {
            const int nChars = 256;
            int handle = 0;
            StringBuilder Buff = new StringBuilder(nChars);

            handle = GetForegroundWindow();

            if (GetWindowText(handle, Buff, nChars) > 0)
            {
                return (Buff.ToString());
            }
            else
                return "";
        }

        private bool isDocumentActive()
        {
            bool result = false;

            if (originalFilename != null)
            {
                string title = getActiveWindowTitle();
                string file = new FileInfo(originalFilename).Name;
                if (title.Contains(file.Substring(0, file.IndexOf('.'))))
                    result = true;
            }

            return result;
        }

        private void compareAndSend(String oldFile,String newFile)
        {
            Console.WriteLine(oldFile);
            Console.WriteLine(newFile);
            //do a diff and send out a message to the buddies
            Diff diff = new Diff(newFile, oldFile, true, true);
            string message = "";
            foreach (Diff.Hunk hunk in diff)
            {
                string hunkString = hunk.ToString();

                if (hunkString.Contains("<") || hunkString.Contains(">"))
                {
                    message = message + hunkString + "\n";
                    message = message.Replace("<", "Added: ");
                    message = message.Replace(">", "Removed: ");
                }
            }

            updateVersionTree();
            //for now, broadcast that a new version is available to all online buddies
            msnManager.broadcastMessage("New version available");
            msnManager.broadcastMessage(message);
        }
        #endregion
    }
}
