using System;
using System.Runtime.InteropServices;
using System.Drawing;
using Microsoft.Win32;
using System.ComponentModel;
using System.Windows.Forms;
using System.Data;
using System.Text;

namespace DesktopLib
{
	/// <summary>
	/// Uses external Win32 functions to capture regions of a desktop.
	/// </summary>
	public class Win32
	{
		private Win32() {}

		private const int SRCCOPY = 0x00CC0020;

		// The GetDesktopWindow function returns a handle to the desktop window. The desktop window covers the entire screen. The desktop window is the area on top of which other windows are painted.
		[DllImport("user32.dll", CharSet=CharSet.Ansi, ExactSpelling=true, SetLastError=false)]
		private static extern IntPtr GetDesktopWindow();

		// The BitBlt function performs a bit-block transfer of the color data corresponding to a rectangle of pixels from the specified source device context into a destination device context.
		[DllImport("gdi32.dll", CharSet=CharSet.Ansi, ExactSpelling=true, SetLastError=false)]
		private static extern UInt64 BitBlt
			(IntPtr hDestDC, int x, int y, int nWidth, int nHeight,
			IntPtr hSrcDC, int xSrc, int ySrc, System.Int32 dwRop);

		// The GetWindowDC function retrieves the device context (DC) for the entire window, including title bar, menus, and scroll bars.
		[DllImport("user32.dll", CharSet=CharSet.Ansi, ExactSpelling=true, SetLastError=false)]
		private static extern IntPtr GetWindowDC(IntPtr hwnd);

		// The ReleaseDC function releases a device context (DC), freeing it for use by other applications.
		[DllImport("user32.dll", CharSet=CharSet.Ansi, ExactSpelling=true, SetLastError=false)]
		private static extern int ReleaseDC(IntPtr hwnd, IntPtr dc);

		[ DllImport("user32.dll") ] 
		static extern int GetForegroundWindow(); 

		[ DllImport("user32.dll") ] 
		static extern int GetWindowText(int hWnd, StringBuilder text, int count); 

		[StructLayout(LayoutKind.Sequential)]
		public struct RECT 
		{
			public int left;
			public int top;
			public int right;
			public int bottom;
		}

		[ DllImport("user32.dll") ] 
		static extern int GetWindowRect(int hWnd, ref RECT rect); 
		

		/// <summary>
		/// Returns a segment of the desktop as an image.
		/// Only captures the specified segment; does not do a full screenshot and then crop.
		/// </summary>
		public static Image GetScreenshot(int x, int y, int width, int height)
		{
			Graphics desktop;
			Image desktopImage;
			Rectangle virtualScreen = Screen.PrimaryScreen.Bounds;

			desktopImage = new Bitmap(width, height);
			using(desktop = Graphics.FromImage(desktopImage))
			{
				IntPtr ptrDesktop = desktop.GetHdc();
				IntPtr ptrDesktopWindow = Win32.GetDesktopWindow();
				IntPtr ptrWindowDC = Win32.GetWindowDC(ptrDesktopWindow);

				Win32.BitBlt(ptrDesktop,0,0,width, height, ptrWindowDC, (virtualScreen.X + x), (virtualScreen.Y + y), SRCCOPY);

				Win32.ReleaseDC(ptrDesktopWindow, ptrWindowDC);
				desktop.ReleaseHdc(ptrDesktop);
				
				ptrDesktop = IntPtr.Zero;
				ptrDesktopWindow = IntPtr.Zero;
				ptrWindowDC = IntPtr.Zero;
			}
						
			return desktopImage;
		} 

		public static IntPtr GetActiveWindowHandle()
		{
			const int nChars = 256;
			int handle = 0;
			StringBuilder Buff = new StringBuilder(nChars);
   
			handle = GetForegroundWindow();

			if ( GetWindowText(handle, Buff, nChars) > 0 )
			{
				return new IntPtr(handle);
			}
			return IntPtr.Zero;
		}

		public static string GetActiveWindowTitle()
		{
			const int nChars = 256;
			int handle = 0;
			StringBuilder Buff = new StringBuilder(nChars);
   
			handle = GetForegroundWindow();

			if ( GetWindowText(handle, Buff, nChars) > 0 )
			{
				return Buff.ToString();
			}
			return string.Empty;
		}

		public static string GetWindowText(IntPtr handle)
		{
			const int nChars = 256;
			StringBuilder Buff = new StringBuilder(nChars);
			if (GetWindowText((int)handle, Buff, nChars) > 0)
			{
				return Buff.ToString();
			}
			return string.Empty;
		}

		public static Rectangle GetWindowBounds(IntPtr handle)
		{
			RECT rc = new RECT();
			if (GetWindowRect((int)handle, ref rc) > 0)
			{
				Rectangle r = new Rectangle(rc.left, rc.top, rc.right-rc.left, rc.bottom-rc.top);
				return r;
			}

			return Rectangle.Empty;
		}
	}
}
